import math


def round_half_up(n: float, decimals: float = 0) -> float:
    """This function rounds to the nearest integer number (e.g 2.4 becomes 2.0 and 2.6 becomes 3);
     in case of tie, it rounds up (e.g. 1.5 becomes 2.0 and not 1.0)
    Args:
        n: number to round
        decimals: number of decimal figures that we want to keep; defaults to zero
    Returns:
        rounded_number: input number rounded with the desired decimals
    """
    multiplier = 10 ** decimals

    rounded_number = math.floor(n * multiplier + 0.5) / multiplier

    return rounded_number


def print_running_time(start_time: float, end_time: float, process_name: str):
    """This function takes as input the start and the end time of a process and prints to console the time elapsed for this process
    Args:
        start_time: instant when the timer was started
        end_time: instant when the timer was stopped
        process_name: name of the process
    Returns:
        None
    """
    sentence = str(process_name)  # convert to string whatever the user inputs as third argument
    temp = end_time - start_time  # compute time difference
    hours = temp // 3600  # compute hours
    temp = temp - 3600 * hours  # if hours is not zero, remove equivalent amount of seconds
    minutes = temp // 60  # compute minutes
    seconds = temp - 60 * minutes  # compute minutes
    print('\n%s time: %d hh %d mm %d ss' % (sentence, hours, minutes, seconds))


import os
import SimpleITK as sitk
import nibabel as nib
from typing import Tuple
import numpy as np


def resample_volume(volume_path: str,
                    new_spacing: list,
                    out_path: str,
                    interpolator: int = sitk.sitkLinear) -> Tuple[sitk.Image, nib.Nifti1Image, np.ndarray]:
    """This function resamples the input volume to a specified voxel spacing
    Args:
        volume_path (str): input volume path
        new_spacing (list): desired voxel spacing that we want
        out_path (str): path where we temporarily save the resampled output volume
        interpolator (int): interpolator that we want to use (e.g. 1= NearNeigh., 2=linear, ...)
    Returns:
        resampled_volume_sitk_obj: resampled volume as sitk object
        resampled_volume_nii_obj: resampled volume as nib object
        resampled_volume_nii: resampled volume as numpy array
    """
    volume = sitk.ReadImage(volume_path)  # read volume
    original_size = volume.GetSize()  # extract size
    original_spacing = volume.GetSpacing()  # extract spacing
    new_size = [int(round(osz * ospc / nspc)) for osz, ospc, nspc in zip(original_size, original_spacing, new_spacing)]
    resampled_volume_sitk_obj = sitk.Resample(volume, new_size, sitk.Transform(), interpolator,
                                              volume.GetOrigin(), new_spacing, volume.GetDirection(), 0,
                                              volume.GetPixelID())
    sitk.WriteImage(resampled_volume_sitk_obj, out_path)  # write sitk volume object to disk
    resampled_volume_nii_obj = nib.load(out_path)  # type: nib.Nifti1Image # load volume as nibabel object
    resampled_volume_nii = np.asanyarray(resampled_volume_nii_obj.dataobj)  # type: np.ndarray # convert from nibabel object to np.array
    os.remove(out_path)  # remove volume from disk to save space

    return resampled_volume_sitk_obj, resampled_volume_nii_obj, resampled_volume_nii


def remove_zeros_ijk_from_volume(input_volume: np.ndarray) -> np.ndarray:
    """This function removes all the rows, columns and slices of the input volume that only contain zero values.
    Args:
        input_volume: volume from which we want to remove zeros
    Returns:
        cropped_volume: cropped volume (i.e. input volume with zeros removed)
    """

    def remove_zeros_one_coordinate(input_volume_: np.ndarray, range_spatial_dim: int, spatial_dim: int):
        idxs_nonzero_slices = []  # will the contain the indexes of all the slices that have nonzero values
        for idx in range(range_spatial_dim):  # loop over coordinate
            if spatial_dim == 0:
                one_slice = input_volume_[idx, :, :]
            elif spatial_dim == 1:
                one_slice = input_volume_[:, idx, :]
            elif spatial_dim == 2:
                one_slice = input_volume_[:, :, idx]
            else:
                raise ValueError("spatial_dim can only be 0, 1, or 2. Got {} instead".format(spatial_dim))

            if np.count_nonzero(one_slice) > 0:  # if the slice has some nonzero values
                idxs_nonzero_slices.append(idx)  # save slice index

        # retain only indexes with nonzero values from the two input volumes
        if spatial_dim == 0:
            cropped_volume_ = input_volume_[idxs_nonzero_slices, :, :]
        elif spatial_dim == 1:
            cropped_volume_ = input_volume_[:, idxs_nonzero_slices, :]
        elif spatial_dim == 2:
            cropped_volume_ = input_volume_[:, :, idxs_nonzero_slices]
        else:
            raise ValueError("spatial_dim can only be 0, 1, or 2. Got {} instead".format(spatial_dim))

        return cropped_volume_

    assert len(input_volume.shape) == 3, "The input volume must be 3D"

    # i coordinate
    cropped_volume = remove_zeros_one_coordinate(input_volume, input_volume.shape[0], spatial_dim=0)
    # j coordinate
    cropped_volume = remove_zeros_one_coordinate(cropped_volume, input_volume.shape[1], spatial_dim=1)
    # k coordinate
    cropped_volume = remove_zeros_one_coordinate(cropped_volume, input_volume.shape[2], spatial_dim=2)

    return cropped_volume


def get_axes_orientations_with_nibabel(input_nifti_volume: nib.Nifti1Image) -> tuple:
    """This function returns the axes orientations as a tuple
    Args:
        input_nifti_volume: the input volume for which we want the axes orientations
    Returns:
        orientations: the axes orientations
    """
    aff_mat = input_nifti_volume.affine  # type: np.ndarray # extract affine matrix
    orientations = nib.aff2axcodes(aff_mat)

    return orientations


def get_nibabel_header(input_nifti_volume: nib.Nifti1Image,
                       print_header: bool = False) -> nib.nifti1.Nifti1Header:
    """This function returns the header of the nifti image/volume
    Args:
        input_nifti_volume: the input volume for which we want the header
        print_header: whether to print the header or not; defaults to False
    Returns:
        header (nib.nifti1.Nifti1Header): the axes orientations
    """
    header = input_nifti_volume.header
    if print_header:
        print(header)

    return header


def read_dcm_series(dcm_dir: str) -> sitk.Image:
    """This function reads a dicom series with SimpleITK
    Args:
        dcm_dir: directory where dicom files are stored
    Returns:
        volume_sitk: volume loaded as sitk.Image
    """
    reader = sitk.ImageSeriesReader()  # create reader
    dicom_names = reader.GetGDCMSeriesFileNames(dcm_dir)  # read dicom series
    reader.SetFileNames(dicom_names)

    volume_sitk = reader.Execute()  # extract sitk.Image

    return volume_sitk


def get_sitk_volume_info(path_to_nii_or_dcm: str, print_info: bool = False) -> dict:
    """This function prints basic info of the input volume
    Args:
        path_to_nii_or_dcm: path to volume that we want to explore
        print_info: whether to print the volume info or no; defaults to False
    Returns:
        volume_info: it contains all the main volume information
    """
    if os.path.isdir(path_to_nii_or_dcm):  # if path_to_nii_or_dcm is a directory
        volume_sitk = read_dcm_series(path_to_nii_or_dcm)
    else:  # if instead path_to_nii_or_dcm is a file
        volume_sitk = sitk.ReadImage(path_to_nii_or_dcm)  # read as sitk Image

    volume_info = {"dimensions": volume_sitk.GetDimension(),
                   "size": volume_sitk.GetSize(),
                   "origin": volume_sitk.GetOrigin(),
                   "spacing": volume_sitk.GetSpacing(),
                   "direction": volume_sitk.GetDirection(),
                   "nb_components_per_pixel": volume_sitk.GetNumberOfComponentsPerPixel(),
                   "pixel_type": volume_sitk.GetPixelID(),
                   "pixel_id_type_as_string": volume_sitk.GetPixelIDTypeAsString(),
                   "pixel_id_value": volume_sitk.GetPixelIDValue()
                   }

    if print_info:
        print("Dimensions: {}".format(volume_sitk.GetDimension()))
        print("Size: {}".format(volume_sitk.GetSize()))
        print("Origin: {}".format(volume_sitk.GetOrigin()))
        print("Spacing: {}".format(volume_sitk.GetSpacing()))
        print("Direction cosine matrix: {}".format(volume_sitk.GetDirection()))
        print("Nb. components per pixel {}".format(volume_sitk.GetNumberOfComponentsPerPixel()))
        print("Pixel type: {}".format(volume_sitk.GetPixelID()))
        print("Pixel ID type as string: {}".format(volume_sitk.GetPixelIDTypeAsString()))
        print("Pixel ID value: {}".format(volume_sitk.GetPixelIDValue()))

    return volume_info



from dateutil.parser import parse


def is_date(input_string: str, fuzzy: bool = False) -> bool:
    """This function checks whether in the input string there is a date
    Args:
        input_string: string to check for date
        fuzzy: ignore unknown tokens in string if True
    Returns:
         True: whether the string can be interpreted as a date.
         False: otherwise
    """
    try:
        parse(input_string, fuzzy=fuzzy)
        return True

    except ValueError:
        return False


def keep_only_digits(input_string: str) -> str:
    """This function takes as input a string and returns the same string but only containing digits
    Args:
        input_string: the input string from which we want to remove non-digit characters
    Returns:
        output_string: the output string that only contains digit characters
    """
    numeric_filter = filter(str.isdigit, input_string)
    output_string = "".join(numeric_filter)

    return output_string


def load_txt_file_as_string(path_txt_file: str) -> str:
    """This function loads a txt file and returns its content as a string.
    Args:
        path_txt_file: path to txt file
    Returns:
        content: content of txt file
    """
    file = open(path_txt_file)
    content = file.read()
    file.close()

    return content


def add_leading_zeros(input_string: str, out_len: int) -> str:
    """This function adds leading zeros to the input string. The output string will have length == out_len
    Args:
        input_string: the input string
        out_len: length of output string with leading zeros
    Returns:
        out_string: the initial string but with leading zeros up to out_len characters
    Example:
        >>> s = "13"
        >>> s_out = add_leading_zeros(input_string=s, out_len=4)
        >>> s_out
        '0013'
    """
    out_string = input_string.zfill(out_len)

    return out_string


def remove_leading_zeros(input_string: str) -> str:
    """This function removes the leading zeros from input_string
    Args:
        input_string: the input string
    Returns:
        out_string: the input string without leading zeros
    """
    # use lstrip setting '0' as leading character, by default it would be space
    out_string = input_string.lstrip('0')

    return out_string


def str2bool(v: str) -> bool:
    """This function converts the input string into a boolean
    Args:
        v: input argument
    Returns:
        True: if the input argument is 'yes', 'true', 't', 'y', '1'
        False: if the input argument is 'no', 'false', 'f', 'n', '0'
    Raises:
        ValueError: if the input argument is none of the above
    """
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise ValueError('Boolean value expected.')



import os
import pickle
from collections import Counter
import glob
import random
from typing import Any, Tuple
import operator


def save_list_to_disk_with_pickle(list_to_save: list, out_dir: str, out_filename: str) -> None:
    """This function saves a list to disk
    Args:
        list_to_save: list that we want to save
        out_dir: path to output folder; will be created if not present
        out_filename: output filename
    Returns:
        None
    """
    if not os.path.exists(out_dir):  # if output folder does not exist
        os.makedirs(out_dir)  # create it
    open_file = open(os.path.join(out_dir, out_filename), "wb")
    pickle.dump(list_to_save, open_file)  # save list with pickle
    open_file.close()


def load_list_from_disk_with_pickle(path_to_list: str) -> list:
    """This function loads a list from disk
    Args:
        path_to_list: path to where the list is saved
    Returns:
        loaded_list: loaded list
    Raises:
        AssertionError: if list path does not exist
    """
    assert os.path.exists(path_to_list), "Path {} does not exist".format(path_to_list)
    open_file = open(path_to_list, "rb")
    loaded_list = pickle.load(open_file)  # load from disk
    open_file.close()

    return loaded_list


def load_list_from_partial_name_with_glob(input_dir: str, partial_filename: str) -> list:
    """This function loads a list from disk by knowing only part of the filename
    Args:
        input_dir: directory where list was saved
        partial_filename: partial filename (use * as wildcare)
    Returns:
        list_of_interest: loaded list
    Example:
        # suppose the filename is y_true_fold_1, we can call:
        >>> y_true = load_list_with_glob(path_to_dir, 'y_true*')
    """
    file_path = glob.glob(os.path.join(input_dir, partial_filename))  # type: list
    assert len(file_path) == 1, "We expect only one filename to match"
    list_of_interest = load_list_from_disk_with_pickle(file_path[0])

    return list_of_interest


def find_common_elements(list1: list, list2: list) -> list:
    """This function takes as input two lists and returns a list with the common elements
    Args:
        list1: first list
        list2: second list
    Returns:
        intersection_as_list: list containing the common elements between the two input lists
    """
    list1_as_set = set(list1)  # type: set
    intersection = list1_as_set.intersection(list2)  # type: set
    intersection_as_list = list(intersection)  # type: list

    return intersection_as_list


def most_frequent_n_elements(input_list: list, n: int) -> Any:
    """This function is given a list as input and it returns its most frequent element
    Args:
        input_list: list where we search the most frequent element
        n: number of most frequent elements we want to retrieve
    Returns:
        most_frequent_item (*): most frequent item in the list; can be of Any type
    """
    occurrence_count = Counter(input_list)  # type: Counter
    most_frequent_n_items = occurrence_count.most_common(n)

    return most_frequent_n_items


def flatten_list(list_of_lists: list) -> list:
    """This function flattens the input list
    Args:
        list_of_lists: input list of lists that we want to flatten
    Returns:
        flattened_list: flattened list
    """
    flattened_list = [item for sublist in list_of_lists for item in sublist]

    return flattened_list


def find_difference_list(list1: list, list2: list) -> list:
    """This function takes as input two lists and returns the difference list between them
    Args:
        list1: first list
        list2: second list
    Returns:
        difference_list: list containing the difference elements between the two input lists
    """
    difference_list = list(set(list1) - set(list2))

    return difference_list


def split_list_equal_sized_groups(lst: list, n: int, seed: int = 123) -> list:
    """This function splits a list in n approximately equal-sized subgroups
    Args:
        lst: input list that we want to split
        n: number of splits
        seed: random seed to use; defaults to 123
    Returns:
        out_list: list of lists, where each internal list is one split
    """
    random.Random(seed).shuffle(lst)  # shuffle list with fixed seed
    division = len(lst) / float(n)  # find number of items per split
    out_list = [lst[int(round(division * i)): int(round(division * (i + 1)))] for i in range(n)]  # list of lists
    return out_list


def find_indexes_where_lists_differ(list1: list, list2: list) -> list:
    """This function returns the indexes where the two input lists differ. The input lists are expected to have same length
    Args:
        list1: first input list
        list2: second input list
    Returns:
        out_list: output list containing the indexes where the two input list differ
    Raises:
        AssertionError: if the two input lists do not have the same length
    """
    assert len(list1) == len(list2), "The two input lists must have same length"
    out_list = [idx for idx, (first, second) in enumerate(zip(list1, list2)) if first != second]
    return out_list


def extract_unique_elements(lst: list, ordered: bool = True) -> list:
    """This function extracts the unique elements of the input list (i.e. it removes duplicates)
    and returns them as an output list; if ordered=True (as by default), the returned list is ordered.
    Args:
        lst: input list from which we want to extract the unique elements
        ordered: whether the output list of unique values is sorted or not; True by default
    Returns:
        out_list: list containing unique values
    """
    out_list = list(set(lst))  # type: list

    if ordered:  # if we want to sort the list of unique values
        out_list.sort()  # type: list

    return out_list


def find_idxs_of_element_in_list(lst: list, element: Any) -> list:
    """This function returns the indexes of the input list that have value == element
    Args:
        lst: input list where we search for indexes
        element: element of which we want to find the indexes
    Returns:
        idxs: list of indexes corresponding to element
    """
    idxs = [i for i, x in enumerate(lst) if x == element]

    return idxs


def list_is_nested(input_list: list) -> bool:
    """This function checks whether the input_list is nested (i.e. it is a list of lists).
    Args:
        input_list: the input list
    Returns:
        is_nested: True if list is nested, False if it isn’t
    """
    is_nested = any(isinstance(i, list) for i in input_list)

    return is_nested


def list_has_duplicates(input_list: list) -> bool:
    """This function checks whether the input_list contains duplicates or not.
    Args:
        input_list: the input list where we look for duplicates
    Returns:
        has_duplicates: True if list has duplicates, False if it doesn't
    """
    if list_is_nested(input_list):
        input_list = flatten_list(input_list)  # flatten list
    has_duplicates = len(input_list) != len(set(input_list))

    return has_duplicates


def first_argmax(input_list) -> int:
    """This function returns the index of the max value. If there are duplicate max values in input_list,
    the index of the first maximum value found will be returned.
    Args:
        input_list: list for which we want to find the argmax
    Returns:
        idx_max: index corresponding to the maximum value
    """
    idx_max = input_list.index(max(input_list))

    return idx_max


def shuffle_two_lists_with_same_order(x: list, y: list, chosen_seed: int = 123) -> Tuple[list, list]:
    """This function shuffles the two input lists with the same order
    Args:
        x: first input list
        y: second input list
        chosen_seed: seed to set for reproducibility
    Returns:
        shuffled_x: shuffled version of first list
        shuffled_y: shuffled version of second list
    Raises:
        AssertionError: if the two input lists do not have the same length
    """
    assert len(x) == len(y), "The two input lists must have the same length"
    random.seed(chosen_seed)  # set seed for reproducibility
    zipped_x_and_y = list(zip(x, y))  # zip x and y together so we don't lose the order when shuffling
    random.shuffle(zipped_x_and_y)  # shuffle
    shuffled_x, shuffled_y = zip(*zipped_x_and_y)  # unzip into x and y

    return shuffled_x, shuffled_y


def slice_by_index(lst: list, indexes) -> list:
    """Slice list by positional indexes.
    Args:
        lst: list to slice.
        indexes: iterable of 0-based indexes of the list positions to return.
    Returns:
        a new list containing elements of lst on positions specified by indexes.
    Examples:
        >>> slice_by_index([], [])
        []
        >>> slice_by_index([], [0, 1])
        []
        >>> slice_by_index(['a', 'b', 'c'], [])
        []
        >>> slice_by_index(['a', 'b', 'c'], [0, 2])
        ['a', 'c']
        >>> slice_by_index(['a', 'b', 'c'], [0, 1])
        ['a', 'b']
        >>> slice_by_index(['a', 'b', 'c'], [1])
        ['b']
    """
    if not lst or not indexes:
        return []
    slice_ = operator.itemgetter(*indexes)(lst)
    if len(indexes) == 1:
        return [slice_]
    return list(slice_)


def keep_only_duplicates(input_list: list) -> list:
    """This function removes all unique values from input_list and keeps only the duplicates
    Args:
        input_list: list from which we want to remove unique values
    Returns:
        list_only_with_duplicates: output list that only contains the duplicates
    Example:
        >>> l = [1, 2, 2, 3, 3, 3, 4]
        >>> out_list = keep_only_duplicates(l)
        >>> out_list
        [2,3]
    """
    counts = Counter(input_list)
    list_only_with_duplicates = [id for id in counts if counts[id] > 1]

    return list_only_with_duplicates


def check_if_string_is_in_any_item_of_list(input_list: list, match_string: str) -> bool:
    """This function checks whether match_string is in any of the items of input_list;
    if yes, it returns True, otherwise it returns False.
    Args:
       input_list: input list where we search for a match
       match_string: string to match
    Returns:
        list_contains_match_string: True if list contains match_string; False otherwise
    """
    list_contains_match_string = bool([item for item in input_list if (match_string in item)])

    return list_contains_match_string



import numpy as np
from typing import Any


def find_most_frequent_value(input_array: np.ndarray) -> Any:
    """This function finds the most common value in the input numpy array
    Args:
        input_array: input array for which we want to find the most frequent value
    Returns:
        most_frequent_value: most frequent value
    """
    values, counts = np.unique(input_array, return_counts=True)

    idx = np.argmax(counts)
    most_frequent_value = values[idx]  # extract the most frequent element

    return most_frequent_value


def has_nans(input_array: np.ndarray) -> bool:
    """"This function checks whether the input array contains any nan.
    If yes, it returns True, otherwise it returns False.
    Args:
        input_array: input array where we check for nans
    Returns:
        array_has_nan: True if input_array contains nans; False otherwise
    """
    array_sum = np.sum(input_array)
    array_has_nan = np.isnan(array_sum)

    return array_has_nan


def is_binary(input_array: np.ndarray) -> bool:
    """This function checks whether the input array is binary (i.e. contains only 0s and 1s).
    If yes, it returns True, otherwise it returns False.
    Args:
        input_array: input array that we want to inspect
    Returns:
        array_is_binary: True if input_array is binary; False otherwise
    """
    array_is_binary = np.array_equal(input_array, input_array.astype(bool))

    return array_is_binary


def has_values_all_in_range(input_array: np.ndarray, low: int, high: int) -> bool:
    """This function checks whether the input array has values that lie within the range (low, high).
    If yes, it returns True, otherwise it returns False.
    Args:
        input_array: input array that we want to inspect
        low: lower bound
        high: upper bound
    Returns:
        array_is_withing_range: True if input_array has values in range (low, high); False otherwise
    """
    mask_array_bool = (input_array > low) & (input_array < high)

    if False in mask_array_bool:  # if there is just one False value
        array_is_withing_range = False
    else:  # if instead all values are True
        array_is_withing_range = True

    return array_is_withing_range


def pad_image_to_specified_shape(input_img: np.ndarray, desired_x_dim: int, desired_y_dim: int) -> np.ndarray:
    """This function zero-pads input_img up to the specified shape (desired_x_dim, desired_y_dim)
    Args:
        input_img: input image that we want to pad
        desired_x_dim: desired dimension 1
        desired_y_dim: desired dimension 2
    Returns:
        padded_img: output padded image
    """

    assert len(input_img.shape) == 2, "This function is intended for 2D arrays"

    # extract dims of input image
    h = input_img.shape[0]
    w = input_img.shape[1]

    # extract padding width (before and after) for rows
    a = (desired_x_dim - h) // 2
    aa = desired_x_dim - a - h

    # extract padding width (before and after) for cols
    b = (desired_y_dim - w) // 2
    bb = desired_y_dim - b - w

    padded_img = np.pad(input_img, pad_width=((a, aa), (b, bb)), mode='constant', constant_values=0)

    return padded_img
